#include <algorithm>
#include <src/server/file/FileServer.h>
#include <log/LogUtils.h>
#include "FileClient.h"
#include <src/InstanceHandler.h>
#include <experimental/filesystem>
#include <misc/memtracker.h>
#include <misc/base64.h>
#include "src/client/ConnectedClient.h"

using namespace std;
using namespace std::chrono;
using namespace ts::server;
namespace fs = std::experimental::filesystem;

#define BUFFER_SIZE (size_t) 2048
FileClient::FileClient(FileServer* handle, int socketFd) : handle(handle), clientFd(socketFd) {
	memtrack::allocated<FileClient>(this);
    this->last_io_action = system_clock::now();

    int enabled = 1;
	if(setsockopt(socketFd, IPPROTO_TCP, TCP_NODELAY, &enabled, sizeof enabled) < 0)
		logError(LOG_FT, "{} Cant enable TCP no delay for socket {}. Error: {}/{}", this->client_prefix(), socketFd, errno, strerror(errno));

    this->readEvent = event_new(this->handle->ioLoop, socketFd, EV_READ|EV_PERSIST, [](int a, short b, void* c){ ((FileClient*) c)->handleMessageRead(a, b, c); }, this);
    this->writeEvent = event_new(this->handle->ioLoop, socketFd, EV_TIMEOUT | EV_WRITE, [](int a, short b, void* c){ ((FileClient*) c)->handleMessageWrite(a, b, c); }, this);


	this->ssl_handler.direct_process(pipes::PROCESS_DIRECTION_OUT, true);
	this->ssl_handler.direct_process(pipes::PROCESS_DIRECTION_IN, true);
    this->ssl_handler.callback_write(bind(&FileClient::sendRawMessage, this, placeholders::_1));
	this->ssl_handler.callback_data(bind(&FileClient::handle_ssl_message, this, placeholders::_1));
    /*
    this->ssl_handler.callback_data([&](const pipes::buffer_view& msg) {
    	if(this->ftType == FTType::TeaWeb_HTTPS) {
			this->handle_http_message(msg);
    	} else if(this->ftType == FTType::TeaWeb_SSL_WS) {
		    this->ws_handler.process_incoming_data(msg);
    	} else {
    		logError(LOG_FT, "{} Decoded SSL packet, but transfer type isn't SSL!", this->client_prefix());
    	}
    });
     */
    //FIXME init ssl error handler?

	this->ws_handler.direct_process(pipes::PROCESS_DIRECTION_OUT, true);
	this->ws_handler.direct_process(pipes::PROCESS_DIRECTION_IN, true);

	/*
	this->ws_handler.callback_data([&](const pipes::WSMessage& message) {
		if(this->state_transfere == T_INITIALIZE) {
			this->applyKey(message.data.string()); //Got the key :)
		} else if(this->state_transfere == T_TRANSFER) {
			if(this->pendingKey->upload)
				this->uploadWriteBytes(message.data.string());
			else {
				logError(LOG_FT, "{} Invalid write (Just download)", this->client_prefix());
				this->disconnect();
			}
		}
	});
	 */
	this->ws_handler.callback_data(bind(&FileClient::handle_ws_message,this, placeholders::_1));
	this->ws_handler.callback_write([&](const pipes::buffer_view& data) {
		this->ssl_handler.send(data);
	});
	this->ws_handler.on_connect = [&]() {};
	this->ws_handler.on_disconnect = [&](const std::string&) {};
}

FileClient::~FileClient() {
    if(this->thread_flush.joinable()){
		this->state_connection = C_DISCONNECTED;

	    if(this->thread_flush.get_id() != this_thread::get_id())
		    this->thread_flush.join();
	    else
		    this->thread_flush.detach();
    }
	memtrack::freed<FileClient>(this);
}

size_t FileClient::used_bandwidth() {
	auto now = system_clock::now();

	size_t tranfarred_bytes = 0;
	auto timeout = now - seconds(1);
	{
		lock_guard<recursive_mutex> lock(this->bandwidth_lock);
		for(auto it = this->bandwidth.end(); it != this->bandwidth.begin();) {
			--it;

			if((*it)->timestamp < timeout) {
				this->bandwidth.erase(this->bandwidth.begin(), it);
				break;
			}
			tranfarred_bytes += (*it)->length;
		}
	}
	return tranfarred_bytes;
}

std::string FileClient::client_prefix() {
	auto ip_address = net::to_string(this->remoteAddress.sin_addr);
	bool hide_ip = config::server::disable_ip_saving;
	if(!hide_ip) {
		auto client = this->client;
		if(client) {
			auto server = client->getServer();
			if(server) {
				hide_ip = server->disable_ip_saving();
			}
		}
	}
	if(hide_ip)
		ip_address = "X.X.X.X";
	if(this->client) return "[" + to_string(this->client->getServerId()) + "|" + ip_address + ":" + to_string(htons(this->remoteAddress.sin_port)) + "| " + this->client->getDisplayName() + "]";
	return "[0|" + ip_address + ":" + to_string(htons(this->remoteAddress.sin_port)) + "|unconnected]";
}

size_t FileClient::transferred_bytes() {
	return this->bytesHandled;
}

size_t FileClient::remaining_bytes() {
	threads::MutexLock lock(this->tickLock);
	if(!this->pendingKey) return 0;
	return this->pendingKey->size - this->bytesHandled;
}

void FileClient::disconnect(std::chrono::milliseconds timeout) {
	auto own_lock = _this.lock();
	if(!own_lock) return;

	bool apply_flush = timeout.count() > 0;
	debugMessage(LOG_FT, "{} Disconnecting client. Connection state: {} Flush IO: {}", this->client_prefix(), (int) this->state_connection, apply_flush);
	unique_lock<threads::Mutex> tick_lock(this->tickLock);

	if(this->state_connection == C_DISCONNECTED) return; /* we're already disconnected */

	if(apply_flush && this->state_connection == C_DISCONNECTING) return; /* we're already flushing the IO */

	if(this->ftType == FTType::TeaWeb_SSL_WS) {
		this->ws_handler.disconnect(1000, "disconnected");
	};

	this->state_connection = apply_flush ? C_DISCONNECTING : C_DISCONNECTED;
	if(this->readEvent)
		event_del_noblock(this->readEvent);

	if(apply_flush){
		lock_guard flush_lock(this->thread_flush_lock);
		assert(!this->thread_flush.joinable());

		this->thread_flush = std::thread([own_lock, timeout]{
			auto beg = system_clock::now();
			while(own_lock->state_connection == C_DISCONNECTING && beg + timeout > system_clock::now()){
				{
					lock_guard buffer_lock(own_lock->bufferLock);
					if(own_lock->read_queue.empty() && own_lock->write_queue.empty()) break;
				}
				usleep(10 * 1000);
			}
			unique_lock<threads::Mutex> l(own_lock->tickLock);
			if(own_lock->state_connection != C_DISCONNECTING) return;
			own_lock->disconnectFinal(l, true);
		});
		{
			auto native_handle = this->thread_flush.native_handle();
			pthread_setname_np(native_handle, "FileClient flush");
		}
	} else {
		unique_lock flush_lock(this->thread_flush_lock);
		if(this->thread_flush.joinable()) {
			flush_lock.unlock();
			this->thread_flush.join();
			flush_lock.lock();
		}
		disconnectFinal(tick_lock, false);
	}
}

void FileClient::disconnectFinal(unique_lock<threads::Mutex>& l, bool lock_flush_thread) {
    auto ownLock = _this.lock();
	unique_lock l1(this->bufferLock);

	this->state_connection = C_DISCONNECTED;
	{
		unique_lock flush_lock(this->thread_flush_lock, try_to_lock);
		if(flush_lock.owns_lock() || !lock_flush_thread) {
			if(this->thread_flush.joinable()) {
				if(this->thread_flush.get_id() == this_thread::get_id())
					this->thread_flush.detach();
				else
					this->thread_flush.join();
			}
		}
	}

	if(this->readEvent){
		auto event = this->readEvent;
		this->readEvent = nullptr;
		l1.unlock();
		l.unlock();
		event_del_block(event);
		event_free(event);
		l.lock();
		l1.lock();
	}

    if(this->writeEvent){
		auto event = this->writeEvent;
		this->writeEvent = nullptr;
	    l1.unlock();
	    l.unlock();
	    event_del_block(event);
	    event_free(event);
	    l.lock();
	    l1.lock();
    }

    if(this->clientFd > 0) {
	    shutdown(this->clientFd, SHUT_RDWR);
        close(this->clientFd);
    }
    this->clientFd = -1;

	{
		threads::MutexLock l2(this->handle->clientLock);
		auto& clList = this->handle->connectedClients;
		auto elm = find(clList.begin(), clList.end(), _this.lock());
		if(elm != clList.end()) clList.erase(elm);
		else logError(LOG_FT, "{} Invalid ft client list!", this->client_prefix());
	}

	this->read_queue.clear();
	this->write_queue.clear();

    if(this->fstream){
        this->fstream->flush();
        this->fstream->close();
        delete this->fstream;
        this->fstream = nullptr;
    }

    this->pendingKey = nullptr;
}

bool FileClient::tick() {
    lock_guard<threads::Mutex> l(this->tickLock);
    if(this->state_connection == C_DISCONNECTED) return false;

    if(this->state_connection != C_DISCONNECTING) {
	    if(last_io_action.time_since_epoch().count() == 0)
		    last_io_action = system_clock::now();
	    else if(last_io_action + minutes(1) < system_clock::now()) {
		    logMessage(LOG_FT, "{} Timed out after one minute of silence!", this->client_prefix());
		    this->disconnect();
		    return false;
	    }
	}

    /* decode incomming stuff */
    bool flag_reexecute = false;
	{
		/* types which require an extra layer of decode */
		if(this->ftType == FTType::TeaWeb_SSL || this->ftType == FTType::TeaWeb_SSL_HTTP || this->ftType == FTType::TeaWeb_SSL_WS || this->ftType == FTType::TeaWeb_HTTP) {
			pipes::buffer buffer;
			{
				lock_guard buffer_lock(this->bufferLock);
				if(this->read_queue.empty()) {
					flag_reexecute = false;
				} else {
					flag_reexecute = true;

					buffer = this->read_queue.front();
					this->read_queue.pop_front();
				}
			}

			if(flag_reexecute) {
				if(this->ftType == FTType::TeaWeb_HTTP)
					this->handle_http_message(buffer);
				else
					this->ssl_handler.process_incoming_data(buffer);
			}
		} else if(this->ftType == FTType::TeamSpeak) {
			flag_reexecute |= this->handle_ts_message();
		} else if(this->ftType == FTType::Unknown) {
			/* we need at least 16 bytes to detect any type */
			if(this->availableBytes() >= 16) {
				auto header = this->peekBytes(16);
				if(header.find("GET") != -1 || header.find("POST") != -1 || header.find("OPTIONS") != -1) {
					debugMessage(LOG_FT, "{} Using HTTP only!", this->client_prefix());
					this->ftType = FTType::TeaWeb_HTTP;
					return true;
				} else if(pipes::SSL::isSSLHeader(header) && serverInstance->sslManager()->web_ssl_options()) {
					debugMessage(LOG_FT, "{} Encrypting pipe with SSL", this->client_prefix());
					this->ftType = FTType::TeaWeb_SSL;
					this->ssl_handler.initialize(serverInstance->sslManager()->web_ssl_options());
					return true;
				} else {
					debugMessage(LOG_FT, "{} Transferring data with the TS protocol", this->client_prefix());
					this->ftType = FTType::TeamSpeak;
					return true;
				}
			}
		} else {
			logError(LOG_FT, "{} Ticked client with unknown protocol type. Closing connection.", this->client_prefix());
			this->disconnect();
			return false;
		}
	}

	if(this->state_transfer == T_TRANSFER) {
		if(this->pendingKey) {
			/* test for download */
			if(!this->pendingKey->upload) {
				/* lets send some data :) */

				if(!fstream) {
					logError(LOG_FT, "{} Missing file stream! Disconnecting client...", this->client_prefix());
					this->disconnect();
					return false;
				}
				size_t count = 0;
				{
					lock_guard lock(this->bufferLock);
					count = this->write_queue.size();
				}

				if(count <= (1024 * 512) / BUFFER_SIZE){ //Max buffer 500Kb
					if(!fstream->good()){
						logError(LOG_FT, "{} Cant finish file download. File isn't good anymore!", this->client_prefix());
						this->disconnect();
						return false;
					}

					pipes::buffer writeBuffer(BUFFER_SIZE);
					auto read = fstream->readsome((char*) writeBuffer.data_ptr(), writeBuffer.length());
					if(read < 0){
						logError(LOG_FT, "{} Invalid file read. Read {} bytes of max {}. Index {}/{}", this->client_prefix(), read, writeBuffer.length(), this->bytesHandled, this->pendingKey->size);
						this->disconnect();
						return false;
					} else if(read == 0){
						if(this->bytesHandled != this->pendingKey->size){
							logError(LOG_FT, "{} Invalid end of file. Expected {} bytes and attempted to read at {}", this->client_prefix(), this->pendingKey->size, this->bytesHandled);
							this->disconnect();
							return false;
						}
					}

					this->bytesHandled += read;
					this->client->getConnectionStatistics()->logFileTransfareOut(read);
					if(read > 0){
						writeBuffer.resize(read);
						this->sendMessage(writeBuffer);
						flag_reexecute |= true;
					}
				}
			}


			/* lets test if we're done */
			if(this->bytesHandled == this->pendingKey->size){
				auto time = duration_cast<milliseconds>(system_clock::now() - this->connect_timestamp).count();
				logMessage(LOG_FT, "{} File transfer completed. Transferred {} bytes in {} milliseconds. Waiting for disconnect.", this->client_prefix(), this->bytesHandled, time);

				this->close_file_handle();
				this->pendingKey.reset();
				this->state_transfer = T_DONE;
				this->finished_timestamp = system_clock::now();

				/* we expect TS3 to hangup the connection */
				if(this->ftType != FTType::TeamSpeak)
					this->disconnect(seconds(5));
				return false;
			}
		}
	} else if(this->state_transfer == T_DONE) {
		if(this->finished_timestamp + seconds(2) < system_clock::now() && this->state_connection == C_CONNECTED) {
			debugMessage(LOG_FT, "{} Disconnecting client after 2 seconds after finish!", this->client_prefix());
			this->disconnect(seconds(5));
		}
	}

    return flag_reexecute;
}

bool FileClient::applyKey(const string &key) {
    {
        threads::MutexLock lock(this->handle->keylock);
        for(const auto& pkey : this->handle->pendingKeys) //needs a copy
            if(pkey && pkey->key == key){
                this->pendingKey = pkey;
                this->handle->pendingKeys.erase(std::find(this->handle->pendingKeys.begin(), this->handle->pendingKeys.end(), pkey));
                break;
            }
    }
    if(!this->pendingKey){
        logError(LOG_FT, "{} Tried to apply an non existing key! (Key: {})", this->client_prefix(), key);
        this->disconnect();
        return false;
    }

	this->client = this->pendingKey->owner.lock();
	if(!this->client) {
		logError(LOG_FT, "{} Tried connect with an invalid key (client offline)! (Key: {})", this->client_prefix(), key);
		this->disconnect();
		return false;
	}

    debugMessage(LOG_FT, "{} Initialized file transfer for file '{}' (Type: {})", this->client_prefix(), this->pendingKey->targetFile, pendingKey->upload ? "upload" : "download");
	if(pendingKey->offset == 0 && pendingKey->upload) {
		try {
			fs::remove(fs::u8path(pendingKey->targetFile));
		} catch (std::exception& e) {}
	}

    fstream = new std::fstream();
    fstream->open(pendingKey->targetFile, (pendingKey->upload ? ios::out : ios::in) | ios::binary | ios::app);
	if(!*fstream) {
		logError(LOG_FT, "{} Failed to open target file {} for {}", this->client_prefix(), this->pendingKey->targetFile, pendingKey->upload ? "upload" : "download");
		delete fstream;
		this->fstream = nullptr;
		return false;
	}
	fstream->seekg(0, std::ios::beg);
    auto fsize = fstream->tellg();
    fstream->seekg(0, std::ios::end);
    fsize = fstream->tellg() - fsize;
    fstream->seekg(this->pendingKey->offset, std::ios::beg);

	if(!*fstream) {
		logError(LOG_FT, "{} Failed to seek within file {} for {}", this->client_prefix(), this->pendingKey->targetFile, pendingKey->upload ? "upload" : "download");
		delete fstream;
		this->fstream = nullptr;
		return false;
	}
    debugMessage(LOG_FT, "{} Received local file size {}. Target size if {}", this->client_prefix(), fsize, pendingKey->size);
	if(this->state_transfer == T_INITIALIZE)
		this->state_transfer = T_TRANSFER;

	this->connect_timestamp = system_clock::now();
    return true;
}


bool FileClient::uploadWriteBytes(const pipes::buffer_view& message) {
	this->client->getConnectionStatistics()->logFileTransfareIn(message.length());
	if(!fstream->good()){
		logError(LOG_FT, "{} uploadWriteBytes(...) called with invalid fstream!", this->client_prefix());
		return false;
	}

    fstream->write(message.data_ptr<char>(), message.length());
    if(!fstream->good()){
        logError(LOG_FT, "{} Invalid file write! ({})", this->client_prefix(), message.length());
        this->disconnect();
        return false;
    }

    this->bytesHandled += message.length();
    return true;
}

void FileClient::close_file_handle() {
	if(this->fstream) {
		this->fstream->flush();
		this->fstream->close();
		delete this->fstream;
		this->fstream = nullptr;
	}
}

void FileClient::sendMessage(const pipes::buffer_view& message) {
	switch(this->ftType) {
		case FTType::TeamSpeak:
		case FTType::TeaWeb_HTTP:
			this->sendRawMessage(message);
			break;
		case FTType::TeaWeb_SSL:
		case FTType::TeaWeb_SSL_HTTP:
			this->ssl_handler.send(message);
			break;
		case FTType::TeaWeb_SSL_WS:
			this->ws_handler.send({pipes::BINARY, message.own_buffer()});
			break;
		default:
			/* Dont log an error because the timeout disconnect function uses this without knowing which proto */
			__asm__("nop");
			return;
	}
}

void FileClient::handle_ssl_message(const pipes::buffer_view &buffer) {
	if(this->ftType == FTType::TeaWeb_SSL_HTTP)
		this->handle_http_message(buffer);
	else if(this->ftType == FTType::TeaWeb_SSL_WS)
		this->ws_handler.process_incoming_data(buffer);
	else if(this->ftType == FTType::TeaWeb_SSL) { /* lets detect if we have HTTP or WebSocket */
		this->read_buffer += buffer;
		this->handle_http_header();
	}
}

void FileClient::handle_http_header() {
	auto header_end = this->read_buffer.find("\r\n\r\n");
	if(header_end == string::npos) {
		if(this->read_buffer.length() > 1024 * 1024 * 4) {
			this->read_buffer = pipes::buffer{};
			logMessage(LOG_FT, "{} Client tried to fillup server memory. Disconnecting client.", this->client_prefix());
			this->disconnect();
			return;
		}
		return;
	}

	auto raw_request = this->read_buffer.view(0, header_end).string();
	http::HttpRequest request{};
	if(!http::parse_request(raw_request, request)) {
		logError(LOG_FT, "{} Failed to parse HTTP request. Disconnecting client.", this->client_prefix());
		this->disconnect();
		return;
	}

	auto header_upgrade = request.findHeader("Upgrade");
	if(header_upgrade && header_upgrade.values[0] == "websocket") {
		debugMessage(LOG_FT, "{} Received WebSocket upgrade. Upgrading connection.", this->client_prefix(), raw_request);
		if(this->ftType == FTType::TeaWeb_SSL)
			this->ftType = FTType::TeaWeb_SSL_WS;
		else {
			//TODO: WebSocket only!
		}
		this->ws_handler.initialize();
		this->ws_handler.process_incoming_data(this->read_buffer);
		this->read_buffer = pipes::buffer{};
		this->handle->tickFileClient(_this.lock()); /* we require a manual reticking */
		return;
	} else {
		/* process the http request header */
		if(this->ftType == FTType::TeaWeb_SSL)
			this->ftType = FTType::TeaWeb_SSL_HTTP;

		http::HttpResponse response{};
		response.setHeader("Connection", {"close"}); /* close the connection instance, we dont want multiple requests */
		response.setHeader("Access-Control-Allow-Methods", {"GET, POST"});
		response.setHeader("Access-Control-Allow-Origin", {"*"});
		response.setHeader("Access-Control-Allow-Headers", request.findHeader("Access-Control-Request-Headers").values); //access-control-allow-headers
		response.setHeader("Access-Control-Max-Age", {"86400"});
		response.setHeader("Access-Control-Expose-Headers", {"*, Content-Length, X-media-bytes, Content-Disposition"});

		/* test for a preflight request: https://developer.mozilla.org/en-US/docs/Glossary/preflight_request */
		auto request_method = request.findHeader("Access-Control-Request-Method");
		if(request_method) {
			debugMessage(LOG_FT, "{} Received preflight request. Sending close response with allow and let the client reconnect.", this->client_prefix());

			response.code = http::code::_200;

			this->read_buffer = this->read_buffer.range(header_end + 4);
			auto raw_response = response.build();
			this->sendMessage(pipes::buffer_view{raw_response.data(), raw_response.length()});
			this->disconnect(seconds(5)); /* write our response & flush */
			return;
		} else {
			auto transfer_key = request.findHeader("transfer-key");
			auto download_name = request.findHeader("download-name");

			if(!transfer_key) {
				response.code = http::code::code(400, "Bad Request");
				debugMessage(LOG_FT, "{} Received invalid HTTP request (Missing transfer key).", this->client_prefix());
			} else {
				if(!this->applyKey(transfer_key.values[0]) || !this->pendingKey) {
					response.code = http::code::code(404, "Not Found");
					debugMessage(LOG_FT, "{} Received invalid HTTP request (Invalid transfer key: {}).", this->client_prefix(), transfer_key.values[0]);
				} else {
					response.code = http::code::code(200, "OK");
					response.setHeader("Content-Length", {to_string(this->pendingKey->size)});

					if(!this->pendingKey->upload) {
						response.setHeader("Content-type", {"application/octet-stream; "});
						response.setHeader("Content-Transfer-Encoding", {"binary"});
						response.setHeader("Content-Disposition", {"attachment; filename=\"" + http::encode_url(download_name ? download_name.values[0] : "TeaWeb Download") + "\""});

						if(this->pendingKey->size > 1) {
							char header_buffer[64];
							auto read = fstream->readsome(header_buffer, 64);
							if(read > 0)
								response.setHeader("X-media-bytes", {base64::encode(header_buffer, read)});
							fstream->seekg(this->pendingKey->offset, std::ios::beg);
						}
					}
				}
			}

			if(!this->pendingKey || !this->pendingKey->upload) {
				auto raw_response = response.build();
				this->sendMessage(pipes::buffer_view{raw_response.data(), raw_response.length()});
			}
			if(response.code->code != 200) {
				this->disconnect(seconds(5)) /* write our response & flush */;
				return;
			}

			this->http_init = true;
			auto overhead = this->read_buffer.range(header_end + 4);
			this->read_buffer = pipes::buffer{}; /* reset the read buffer */
			if(overhead.length() > 0)
				this->handle_http_message(overhead);
			this->handle->tickFileClient(_this.lock()); /* we require a manual reticking */
		}
	}
}

bool FileClient::handle_ts_message() {
	if(this->state_transfer == T_INITIALIZE) {
		if(this->availableBytes() >= 16) {
			return this->applyKey(this->getBytes(16));
		} else
			return false;
	} else if(this->state_transfer == T_TRANSFER) {
		if(!this->pendingKey)
			return false;

		if(!this->pendingKey->upload)
			return false; /* should never happen! */

		bool reexecute = false;
		pipes::buffer buffer;
		{
			lock_guard buffer_lock(this->bufferLock);
			if(this->read_queue.empty())
				return false; /* nothing to upload */

			buffer = this->read_queue.front();
			this->read_queue.pop_front();
			reexecute |= !this->read_queue.empty();
		}

		if(!this->uploadWriteBytes(buffer))
			return false; /* error already handeled by uploadWriteBytes(...) */

		return reexecute;
	}

	return false;
}

void FileClient::handle_ws_message(const pipes::WSMessage &message) {
	if(this->state_transfer == T_INITIALIZE) {
		debugMessage(LOG_FT, "{} Received transfer key: {}", this->client_prefix(), message.data.string());
		if(this->applyKey(message.data.string()))
			this->handle->tickFileClient(_this.lock()); /* we require a manual reticking */
		return;
	} else if(this->state_transfer == T_TRANSFER) {
		if(this->pendingKey->upload)
			this->uploadWriteBytes(message.data);
		else {
			logError(LOG_FT, "{} Invalid write (Just download)", this->client_prefix());
			this->disconnect();
		}
	}
}

void FileClient::handle_http_message(const pipes::buffer_view &message) {
	if(!this->http_init) {
		this->read_buffer += message;
		this->handle_http_header();
		return;
	}
	if(!https_upload_init) {
		//------WebKitFormBoundaryaWP8XAzMBnMOJznv\r\nContent-Disposition: form-data; name=\"file\"; filename=\"blob\"\r\nContent-Type: application/octet-stream\r\n\r\n
		this->read_buffer += message;

		auto header_end = this->read_buffer.find("\r\n\r\n");
		if(header_end == string::npos) {
			if(this->read_buffer.length() > 1024 * 1024 * 4) {
				this->read_buffer = pipes::buffer{};
				logMessage(LOG_FT, "{} Client tried to fillup server memory. Disconnecting client.", this->client_prefix());
				this->disconnect();
				return;
			}
			return;
		}

		https_upload_init = true;
		auto overhead = this->read_buffer.view(header_end + 4);
		this->read_buffer = pipes::buffer{};
		if(!overhead.empty())
			this->handle_http_message(overhead);
		return;
	}
	if(!this->pendingKey || !this->pendingKey->upload) {
		logError(LOG_FT, "{} HTTP Invalid request", this->client_prefix());
		return;
	}

	auto bytes_to_write = this->pendingKey->size - this->bytesHandled; /* ignore boundaries */
	if(bytes_to_write < message.length())
		this->uploadWriteBytes(message.view(0, bytes_to_write));
	else
		this->uploadWriteBytes(message);

	if(this->bytesHandled == this->pendingKey->size){
		http::HttpResponse response{};
		response.setHeader("Connection", {"close"}); /* close the connection instance, we dont want multiple requests */
		response.setHeader("Access-Control-Allow-Methods", {"GET, POST"});
		response.setHeader("Access-Control-Allow-Origin", {"*"});
		response.setHeader("Access-Control-Allow-Headers", {"*"});
		response.setHeader("Access-Control-Max-Age", {"86400"});
		response.setHeader("Access-Control-Expose-Headers", {"*"});

		auto raw_response = response.build();
		this->sendMessage(pipes::buffer_view{raw_response.data(), raw_response.length()});
	}
}