#include "PlayablePlaylist.h"
#include "src/TSServer.h"
#include "src/client/music/MusicClient.h"
#include "src/client/ConnectedClient.h"
#include <log/LogUtils.h>

using namespace ts;
using namespace ts::server;
using namespace ts::music;
using namespace std;
using namespace std::chrono;

PlayablePlaylist::PlayablePlaylist(const std::shared_ptr<MusicBotManager> &handle, const std::shared_ptr<Properties> &props, const std::shared_ptr<permission::PermissionManager>& perms) : Playlist(handle, props, perms) { }

PlayablePlaylist::~PlayablePlaylist() {}

void PlayablePlaylist::set_self_ref(const std::shared_ptr<Playlist> &ptr) {
	Playlist::set_self_ref(ptr);
}

std::shared_ptr<ts::music::PlayableSong> PlayablePlaylist::pop_next() {
	unique_lock lock(this->currently_playing_lock);

	auto entry = this->loading_entry.lock();
	if(!entry) {
		lock.unlock();
		entry = this->pop_next_entry();
		lock.lock();
		this->loading_entry = entry;

		auto id = entry ? entry->id : 0;
		if(this->properties()[property::PLAYLIST_CURRENT_SONG_ID].as<SongId>() != id)
			this->properties()[property::PLAYLIST_CURRENT_SONG_ID] = id;
	}
	if(!entry) return nullptr;

	if(!entry->load_future) {
		this->loading_entry.reset();
		return nullptr; /* skip this entry */
	}

	if(entry->load_future->state() == threads::FutureState::WORKING) {
		if(entry->load_start + seconds(30) < system_clock::now()) {
			this->loading_entry.reset();
			logError(this->get_server_id(), "[PlayList] Failed to load song info. Error: {}. Timeout. Removing song", entry->load_future->errorMegssage());
			this->delete_song(entry->id); /* acquired playlist */

			return PlayableSong::create({
                    entry->id,
                    entry->url,
                    entry->invoker
            }, MusicClient::failedLoader("Failed to parse song info. Info loader timeouted!"));
		}

		return nullptr;
	}

	this->loading_entry.reset(); /* reset the loading entry because we're returning it when succeeded, else we skip*/
	if(!entry->load_future->succeeded() || !entry->load_future->getValue(nullptr)) {
		//TODO log error within channel?
		logError(this->get_server_id(), "[PlayList] Failed to load song info. Error: {}. Removing song", entry->load_future->errorMegssage());
		this->delete_song(entry->id); /* acquired playlist */

		/* return promise */
		return PlayableSong::create({
                entry->id,
                entry->url,
                entry->invoker
        }, MusicClient::failedLoader("Failed to parse song info: " + entry->load_future->errorMegssage()));
	}

	auto result = entry->load_future->getValue(nullptr);
	assert(result); /* previous tested */

	if(result->type != ::music::UrlType::TYPE_VIDEO && result->type != ::music::UrlType::TYPE_STREAM) {
		/* we cant replay that kind of entry. Skipping it */
		debugMessage(this->get_server_id(), "[PlayList] Tried to replay an invalid entry type (SongID: {}, Type: {})! Skipping song", entry->id, result->type);
		return nullptr;
	}

	return PlayableSong::create({
            entry->id,
            result->url,
            entry->invoker
	}, MusicClient::providerLoader(entry->url_loader));
}

std::shared_ptr<PlaylistEntryInfo> PlayablePlaylist::pop_next_entry() {
	if(!this->playlist_head) return nullptr; /* fuzzy check if we're not maybe empty */

	auto replay_mode = this->properties()[property::PLAYLIST_REPLAY_MODE].as<ReplayMode::value>();
	auto songs = this->list_songs();
	if(songs.empty()) return nullptr; /* we've nothing to play */

	unique_lock playlist_lock(this->playlist_lock);
	auto current_song = this->playlist_find(playlist_lock, this->currently_playing());

	if(replay_mode == ReplayMode::SINGLE_LOOPED) {
		if(current_song) return current_song->entry;
		return songs.front();
	}

	std::shared_ptr<PlaylistEntryInfo> result;
	if(replay_mode == ReplayMode::LINEAR || replay_mode == ReplayMode::LINEAR_LOOPED) {
		if(current_song) {
			if(current_song->next_song)
				result = current_song->next_song->entry;
			else if(replay_mode == ReplayMode::LINEAR_LOOPED && this->playlist_head)
				result = this->playlist_head->entry;
			else {
				this->properties()[property::PLAYLIST_FLAG_FINISHED] = true;
				result = nullptr;
			}
		} else if(replay_mode == ReplayMode::LINEAR_LOOPED || !this->properties()[property::PLAYLIST_FLAG_FINISHED]) {
			result = songs.front();
		}
	} else if(replay_mode == ReplayMode::SHUFFLE) {
		//TODO may add a already played list?
		if(songs.size() == 1) {
			this->properties()[property::PLAYLIST_FLAG_FINISHED] = true;
			result = nullptr;
		} else {
			size_t index;
			while(songs[index = (rand() % songs.size())] == (!current_song ? nullptr : current_song->entry)) { }

			result = songs[index];
		}
	}
	playlist_lock.unlock();

	if(current_song && this->properties()[property::PLAYLIST_FLAG_DELETE_PLAYED].as<bool>())
		this->delete_song(current_song->entry->id);
	return result;
}

bool PlayablePlaylist::set_current_song(SongId song_id) {
	{
		unique_lock playlist_lock(this->playlist_lock);
		auto current_song = this->playlist_find(playlist_lock, song_id);
		this->loading_entry = current_song->entry;

		if(current_song) {
			this->enqueue_load(current_song->entry);
			auto id = current_song ? current_song->entry->id : 0;
			if(this->properties()[property::PLAYLIST_CURRENT_SONG_ID].as<SongId>() != id)
				this->properties()[property::PLAYLIST_CURRENT_SONG_ID] = id;
		}
	}
	{
		auto bot = this->current_bot();
		bot->forwardSong(); /* skip to the song */
	}

	return true;
}