//
// Created by wolverindev on 22.10.17.
//

#include "QueryServer.h"
#include <algorithm>
#include <netinet/tcp.h>
#include <poll.h>
#include <src/TSServer.h>
#include <src/client/query/QueryClient.h>
#include <src/client/InternalClient.h>
#include <misc/rnd.h>
#include <src/InstanceHandler.h>
#include <log/LogUtils.h>
#include <src/client/ConnectedClient.h>

using namespace std;
using namespace std::chrono;
using namespace ts;
using namespace ts::server;

#if defined(TCP_CORK) && !defined(TCP_NOPUSH)
    #define TCP_NOPUSH TCP_CORK
#endif

QueryServer::QueryServer(sql::SqlManager* db) : sql(db) {
	this->_executePool = new threads::ThreadPool(4, "EXEC Query");
}

QueryServer::~QueryServer() {
    stop();
	this->_executePool->shutdown();
	delete this->_executePool;
}


void QueryServer::unregisterConnection(const shared_ptr<QueryClient> &client) {
	{
		threads::MutexLock lock(this->clientLock);
		auto found = std::find(this->connectedClients.begin(), this->connectedClients.end(), client);
		if(found != this->connectedClients.end())
			this->connectedClients.erase(found);
		else
			logError(LOG_QUERY, "Attempted to unregister an invalid query connection!");
	}

    if(client->server) {
	    client->server->getGroupManager()->disableCache(client->getClientDatabaseId());
    } else {
    	serverInstance->getGroupManager()->disableCache(client->getClientDatabaseId());
    }
    /* client->handle = nullptr; */
}

bool QueryServer::start(const sockaddr_in& localAdress, std::string& errorMessage) {
    if(this->running()) return false;
    this->active = true;
    boundAddress = new sockaddr_in;
    memcpy(boundAddress, &localAdress, sizeof(localAdress));

    ip_blacklist.reset(new IpListManager("query_ip_blacklist.txt", {"#A new line separated address blacklist", "#", "#For example if we dont want google:", "8.8.8.8"}));
	ip_whitelist.reset(new IpListManager("query_ip_whitelist.txt", {"#A new line separated address whitelist", "#Every ip have no flood and login attempt limit!", "127.0.0.1/8", "::1"}));
	string error;
	if(!this->ip_blacklist->reload(error)) logError(LOG_QUERY, "Failed to load query blacklist: {}", error);
	if(!this->ip_whitelist->reload(error)) logError(LOG_QUERY, "Failed to load query whitelist: {}", error);
    this->serverSocket = socket(AF_INET, SOCK_STREAM | SOCK_NONBLOCK, 0);
    if (serverSocket < 0) {
        logCritical("Cant create server socket for file server");
        return false;
    }

    int enable = 1;
    int disabled = 0;
    if (setsockopt(serverSocket, SOL_SOCKET, SO_REUSEADDR, &enable, sizeof(int)) < 0)
        logError("setsockopt(SO_REUSEADDR) failed");
    if(setsockopt(serverSocket, IPPROTO_TCP, TCP_NOPUSH, &disabled, sizeof disabled) < 0)
        logError("Cant disable nopush! Error: "+to_string(errno)+" / "+strerror(errno));
    setsockopt(serverSocket, SOL_SOCKET, SO_KEEPALIVE, &enable, sizeof(enable));

    if(fcntl(serverSocket, F_SETFD, FD_CLOEXEC) < 0)
        logError(LOG_QUERY, "Failed to enable FD_CLOEXEC for {} (QueryServer)", serverSocket);

    if (bind(serverSocket, (struct sockaddr *) &localAdress, sizeof(localAdress)) < 0) {
        errorMessage = string() + "Cant bind server socket (" + strerror(errno) + ")";
        return false;
    }
    if(listen(serverSocket, 255) < 0){
        errorMessage = string() + "Cant listen on server socket (" + strerror(errno) + ")";
        return false;
    }

    this->eventLoop = event_base_new();
    this->acceptEvent = event_new(this->eventLoop, this->serverSocket, EV_READ | EV_PERSIST, [](int a, short b, void* c){ ((QueryServer*) c)->onClientAccept(a, b, c); }, this);
    event_add(this->acceptEvent, nullptr);

    this->ioThread = new threads::Thread(THREAD_SAVE_OPERATIONS | THREAD_EXECUTE_LATER, [&](){
        debugMessage(LOG_QUERY, "Event base executed ({})", (void*) this->eventLoop);
        event_base_dispatch(this->eventLoop);
        debugMessage(LOG_QUERY, "Event base terminated ({})", (void*) this->eventLoop);
    });
    this->ioThread->name("EVENT Query").execute();

    this->tickingId = serverInstance->scheduler()->schedule("query", bind(&QueryServer::tick, this), seconds(1));
    return true;
}

void QueryServer::stop() {
    if(!this->running()) return;
    active = false;
	serverInstance->scheduler()->cancelTask("query");

    this->clientLock.lock();
    auto clList = this->connectedClients;
    this->clientLock.unlock();

    Command cmd("serverstop");
    cmd["stopped"] = true;
    for(const auto &client : clList){
        client->sendCommand(cmd);
        client->disconnect("server stopped");
    }

    if(this->acceptEvent){
        event_del(this->acceptEvent);
        event_free(this->acceptEvent);
        this->acceptEvent = nullptr;
    }

    {
        auto now = system_clock::now();
        while(!this->connectedClients.empty()) {
            if(now + seconds(5) < system_clock::now()) {
                logError(LOG_QUERY, "Failed to disconnect all clients!");
                break;
            }
            threads::self::sleep_for(milliseconds(100));
        }
    }

    for(auto thread : this->threads){
        if(thread->state() == threads::ThreadState::RUNNING)
            thread->join();
        delete thread;
    }
    this->threads.clear();

	if(this->eventLoop)
		event_base_loopexit(this->eventLoop, nullptr);
    if(this->ioThread) {
	    if(this->ioThread->join(seconds(3)) != 0) {
		    logCritical(LOG_QUERY, "Failed to terminate event loop!");
		    this->ioThread->detach();
	    }
    }
    delete this->ioThread;
    this->ioThread = nullptr;

	if(this->eventLoop) {
		event_base_free(this->eventLoop);
		this->eventLoop = nullptr;
	}

    delete this->boundAddress;
    this->boundAddress = nullptr;

    if(this->serverSocket > 0) {
        if(shutdown(this->serverSocket, SHUT_RDWR) < 0) logError(LOG_QUERY, "Could not shutdown server socket!");
        if(close(this->serverSocket) < 0) logError(LOG_QUERY, "Could not close server socket!");
    }
    this->serverSocket = -1;
}

void QueryServer::onClientAccept(int fd, short ev, void *arg) {
    sockaddr_in remoteAddr{};
    memset(&remoteAddr, 0, sizeof(sockaddr_in));
    socklen_t addrLength = sizeof(remoteAddr);

    int acceptedSocketFd = accept(serverSocket, (struct sockaddr *) &remoteAddr, &addrLength);
    if (acceptedSocketFd < 0) {
        if(errno == EAGAIN) { //No manager
            return;
        }
	    logMessage(LOG_QUERY, "Got an error while accepting a new client. (errno: {}, message: {})", errno, strerror(errno));
        return;
    }

    shared_ptr<QueryClient> client = std::make_shared<QueryClient>(this, acceptedSocketFd);
    client->applySelfLock(client);
    memcpy(&client->remote_address, &remoteAddr, sizeof(sockaddr_in));

    this->clientLock.lock();
    this->connectedClients.push_back(client);
    this->clientLock.unlock();
    client->preInitialize();
    if(client->readEvent)
    	event_add(client->readEvent, nullptr);
    logMessage(LOG_QUERY, "Got new client from {}", client->getLoggingPeerIp() + ":" + to_string(client->getPeerPort()));
}

/*
#define QUERY_PASSWORD_LENGTH 7
std::string QueryServer::resetQueryPassword(const std::shared_ptr<QueryLoginCredentials>& credits, std::string password) {
    if(password.empty()) password = rnd_string(QUERY_PASSWORD_LENGTH);
    auto password_copy = password;
    auto res = sql::command(this->sql, "UPDATE `queries` SET `password` = :password WHERE `uniqueId` = :uid AND `username` = :name", variable{":name", credits->username}, variable{":uid", credits->uniqueId}, variable{":password", password}).execute();
    LOG_SQL_CMD(res);
    return password_copy; //Analize why I have to copy that shit here?
}

std::string QueryServer::createQueryLogin(const string &name, ClientUid uid, string password) {
    bool exists = false;
    sql::command(this->sql, "SELECT * FROM `queries` WHERE `username` = :name", variable{":name", name}).query([](bool* flag, int, char**, char**){
        *flag = true;
        return 0;
    }, &exists);
    if(exists) return "";

    if(password.empty())
        password = rnd_string(QUERY_PASSWORD_LENGTH);
    sql::command(this->sql, "INSERT INTO `queries` (`username`, `password`, `uniqueId`) VALUES (:name, :password, :uid)", variable{":name", name}, variable{":uid", uid}, variable{":password", password}).execute();

    return password;
}

bool QueryServer::renameQueryLogin(ClientUid uid, const string &name) {
    bool exists = false;
    sql::command(this->sql, "SELECT * FROM `queries` WHERE `username` = :name", variable{":name", name}).query([](bool* flag, int, char**, char**){
        *flag = true;
        return 0;
    }, &exists);
    if(!exists) return false;

    auto res = sql::command(this->sql, "UPDATE `queries` SET `username` = :name WHERE `uniqueId` = :uid", variable{":name", name}, variable{":uid", uid}).execute();
    LOG_SQL_CMD(res);

    return res;
}

std::shared_ptr<QueryLoginCredentials> QueryServer::findQueryLoginByName(const string &name) {
    std::shared_ptr<QueryLoginCredentials> result = std::make_shared<QueryLoginCredentials>();
    sql::command(this->sql, "SELECT * FROM `queries` WHERE `username` = :name", variable{":name", name}).query([](QueryLoginCredentials* res, int length, char** value, char** columns){
        for(int index = 0; index < length; index++)
            if(strcmp(columns[index], "username") == 0)
                res->username = value[index];
            else if(strcmp(columns[index], "password") == 0)
                res->password = value[index];
            else if(strcmp(columns[index], "uniqueId") == 0)
                res->uniqueId = value[index];
        return 0;
    }, result.get());
    if(result->username.empty()) return nullptr;
    return result;
}

std::shared_ptr<QueryLoginCredentials> QueryServer::findQueryLoginByUid(const string &uid) {
    std::shared_ptr<QueryLoginCredentials> result = std::make_shared<QueryLoginCredentials>();
    sql::command(this->sql, "SELECT * FROM `queries` WHERE `uniqueId` = :name", variable{":name", uid}).query([](QueryLoginCredentials *res, int length, char **value, char **columns) {
        for (int index = 0; index < length; index++)
            if (strcmp(columns[index], "username") == 0)
                res->username = value[index];
            else if (strcmp(columns[index], "password") == 0)
                res->password = value[index];
            else if (strcmp(columns[index], "uniqueId") == 0)
                res->uniqueId = value[index];
        return 0;
    }, result.get());
    if (result->username.empty()) return nullptr;

    return result;
}
 */

/* api */
inline deque<shared_ptr<QueryAccount>> query_accounts(sql::command& command) {
    deque<shared_ptr<QueryAccount>> result;
    command.query([&](int length, std::string* value, std::string* columns){
        auto entry = std::make_shared<PasswortedQueryAccount>();
        for(int index = 0; index < length; index++){
            try {
                if(columns[index] == "username")
                    entry->username = value[index];
                else if(columns[index] == "password")
                    entry->password = value[index];
                else if(columns[index] == "uniqueId")
                    entry->unique_id = value[index];
                else if(columns[index] == "server") {
	                entry->bound_server = value[index].empty() ? 0 : stoll(value[index]);
                }
            } catch (std::exception& ex) {
                logError(LOG_QUERY, "Failed to parse query account data for row {} ({})", columns[index], value[index]);
                return 0;
            }
        }

        result.push_back(entry);
        return 0;
    });

    return result;
}

std::shared_ptr<QueryAccount> QueryServer::create_query_account(const std::string &username, ts::ServerId server, const std::string &owner, const std::string &password) {
	LOG_SQL_CMD(sql::command(this->sql, "INSERT INTO `queries` (`username`, `password`, `uniqueId`, `server`) VALUES (:name, :password, :uid, :server)", variable{":name", username}, variable{":uid", owner}, variable{":password", password}, variable{":server", server}).execute());

    return this->find_query_account_by_name(username);
}

std::shared_ptr<PasswortedQueryAccount> QueryServer::load_password(const std::shared_ptr<QueryAccount> &account) {
    return dynamic_pointer_cast<PasswortedQueryAccount>(account);
}

bool QueryServer::delete_query_account(const std::shared_ptr<ts::server::QueryAccount> &account) {
    LOG_SQL_CMD(sql::command(this->sql, "DELETE FROM `queries` WHERE `username` = :username AND `server` = :server", variable{":username", account->username}, variable{":server", account->bound_server}).execute());

    return true;
}

std::deque<std::shared_ptr<QueryAccount>> QueryServer::list_query_accounts(OptionalServerId server_id) {
    auto command = (
            server_id != EmptyServerId ?
            sql::command(this->sql, "SELECT * FROM `queries` WHERE `server` = :server", variable{":server", server_id}) :
            sql::command(this->sql, "SELECT * FROM `queries`")
    );
    auto accounts = query_accounts(command);
    return accounts;
}

std::shared_ptr<QueryAccount> QueryServer::find_query_account_by_name(const std::string &name) {
    auto command = sql::command(this->sql, "SELECT * FROM `queries` WHERE `username` = :name", variable{":name", name});
    auto accounts = query_accounts(command);
    if(accounts.empty()) return nullptr;
    return accounts.back();
}

deque<shared_ptr<QueryAccount>> QueryServer::find_query_accounts_by_unique_id(const std::string &unique_id) {
    auto command = sql::command(this->sql, "SELECT * FROM `queries` WHERE `uniqueId` = :unique_id", variable{":name", unique_id});
    return query_accounts(command);
}

bool QueryServer::rename_query_account(const std::shared_ptr<ts::server::QueryAccount> &account, const std::string &new_name) {
	LOG_SQL_CMD(sql::command(this->sql, "UPDATE `queries` SET `username` = :new_name WHERE `username` = :old_name AND `server` = :server", variable{":new_name",new_name}, variable{":old_name", account->username}, variable{":server", account->bound_server}).execute());
    return true;
}

bool QueryServer::change_query_password(const std::shared_ptr<ts::server::QueryAccount> &account, const std::string &password) {
	LOG_SQL_CMD(sql::command(this->sql, "UPDATE `queries` SET `password` = :password WHERE `username` = :name AND `server` = :server", variable{":password", password}, variable{":name", account->username}, variable{":server", account->bound_server}).execute());
    return true;
}

void QueryServer::tick() {

    decltype(this->connectedClients) clCopy;
    {
        threads::MutexLock lock(this->clientLock);
        clCopy = this->connectedClients;
    }
    for(const auto& cl : clCopy) cl->queryTick();
    {
        threads::MutexLock lock(this->loginLock);

        vector<std::string> qbanErase;
        for(auto& elm : this->queryBann) {
            if(elm.second < system_clock::now())
                qbanErase.push_back(elm.first);
        }
        for(const auto& ip : qbanErase)
            this->queryBann.erase(ip);

        if(system_clock::now() - seconds(5) < lastDecrease) {
            this->lastDecrease = system_clock::now();

            vector<std::string> lattempErase;
            for(auto& elm : this->loginAttempts) {
                if(elm.second == 0)
                    lattempErase.push_back(elm.first);
                else
                    elm.second--;
            }
            for(const auto& ip : qbanErase)
                this->loginAttempts.erase(ip);
        }
    }
}