#include <deque>
#include <log/LogUtils.h>
#include "WebServer.h"
#include "src/client/web/WebClient.h"
#include <netinet/tcp.h>
#include <src/client/ConnectedClient.h>
#include "src/InstanceHandler.h"

using namespace std;
using namespace std::chrono;
using namespace ts;
using namespace ts::server;

#if defined(TCP_CORK) && !defined(TCP_NOPUSH)
    #define TCP_NOPUSH TCP_CORK
#endif

WebControlServer::WebControlServer(const std::shared_ptr<TSServer>& handle) : handle(handle) {}
WebControlServer::~WebControlServer() = default;

bool WebControlServer::start(sockaddr_in* addr, std::string& errorMessage) {
    if(this->running()) return false;
    this->_running = true;
    boundAddress = new sockaddr_in;
    memcpy(boundAddress, addr, sizeof(*addr));

    this->serverSocket = socket(AF_INET, SOCK_STREAM | SOCK_NONBLOCK, 0);
    if (serverSocket < 0) {
	    logCritical(this->handle->getServerId(), "Cant create server socket for server");
	    return false;
    }

    int enable = 1;
    int disabled = 0;
    if (setsockopt(serverSocket, SOL_SOCKET, SO_REUSEADDR, &enable, sizeof(int)) < 0)
        logError("setsockopt(SO_REUSEADDR) failed");
    if(setsockopt(serverSocket, IPPROTO_TCP, TCP_NOPUSH, &disabled, sizeof disabled) < 0)
        logError("Cant disable nopush! Error: "+to_string(errno)+" / "+strerror(errno));
    setsockopt(serverSocket, SOL_SOCKET, SO_KEEPALIVE, &enable, sizeof(enable));

    if(fcntl(serverSocket, F_SETFD, FD_CLOEXEC) < 0)
        logError(this->handle->getServerId(), "Failed to enable FD_CLOEXEC for {} (WebControlServer)", serverSocket);

    if (bind(serverSocket, (struct sockaddr *) boundAddress, sizeof(*boundAddress)) < 0) {
        errorMessage = string() + "Cant bind server socket (" + strerror(errno) + ")";
        return false;
    }
    if(listen(serverSocket, 255) < 0){
        errorMessage = string() + "Cant listen on server socket (" + strerror(errno) + ")";
        return false;
    }

    auto io_base = serverInstance->getWebIoLoop()->next_loop();
    assert(io_base);
    this->acceptEvent = event_new(io_base->loop, this->serverSocket, EV_READ | EV_PERSIST, [](int a, short b, void* c){ ((WebControlServer*) c)->onClientAccept(a, b, c); }, this);
    event_add(this->acceptEvent, nullptr);

    return true;
}

void WebControlServer::onClientAccept(int fd, short ev, void *arg) {
    sockaddr_in remoteAddr{};
    memset(&remoteAddr, 0, sizeof(sockaddr_in));
    socklen_t addrLength = sizeof(remoteAddr);

    int acceptedSocketFd = accept(this->serverSocket, (struct sockaddr *) &remoteAddr, &addrLength);
    if (acceptedSocketFd < 0) {
        if(errno == EAGAIN) //No manager
            return;

        logError(this->handle->getServerId(), lstream << "Having an error while accepting a new manager. ("+to_string(errno)+"/"+string(strerror(errno))+")");
        return;
    }

    shared_ptr<WebClient> client = std::make_shared<WebClient>(this, acceptedSocketFd, remoteAddr);
    client->applySelfLock(client);
    client->initialize();
    memcpy(&client->remote_address, &remoteAddr, sizeof(sockaddr_in));

    this->clientLock.lock();
    this->clients.push_back(client);
    this->clientLock.unlock();
    event_add(client->readEvent, nullptr);
    logMessage(this->handle->getServerId(), lstream << "[Web] Got new manager from " << client->getLoggingPeerIp() << ":" << client->getPeerPort() << endl);
}

void WebControlServer::stop() {
    //TODO
    this->clientLock.lock();
    auto clList = this->clients;
    this->clientLock.unlock();

    for(const auto& e : clList) {
	    e->closeConnection(system_clock::now());
    }

	if(this->acceptEvent) {
		event_del(this->acceptEvent);
		event_free(this->acceptEvent);
	}

    if(this->serverSocket > 0){
        if(shutdown(this->serverSocket, SHUT_RDWR) < 0) logError(this->handle->getServerId(), "Could not shutdown WebSocketServer socket!");
        if(close(this->serverSocket) < 0) logError(this->handle->getServerId(), "Could not close WebSocketServer socket!");
    }
    this->serverSocket = -1;

    delete this->boundAddress;
    this->boundAddress = nullptr;
}

void WebControlServer::unregisterConnection(const std::shared_ptr<WebClient>& connection) {
    //TODO may checks?
    {
        threads::MutexLock l(this->clientLock);
        for(const auto& cl : this->clients) {
            if(cl == connection) {
                clients.erase(std::find(this->clients.begin(), this->clients.end(), cl));
                break;
            }
        }
    }
}
