#include "FileServer.h"
#include "src/client/file/FileClient.h"
#include "src/client/ConnectedClient.h"
#include <netinet/tcp.h>
#include <experimental/filesystem>
#include <log/LogUtils.h>
#include <misc/std_unique_ptr.h>
#include "src/InstanceHandler.h"

using namespace std;
using namespace std::chrono;
using namespace ts;
using namespace ts::file;
using namespace ts::server;
namespace fs = std::experimental::filesystem;

#if defined(TCP_CORK) && !defined(TCP_NOPUSH)
    #define TCP_NOPUSH TCP_CORK
#endif

extern InstanceHandler* serverInstance;

FileServer::FileServer() {}

FileServer::~FileServer() {
    stop();
}

inline fs::path buildPath(std::string rootPath, std::shared_ptr<file::FileEntry> directory){
    auto strPath = directory ? directory->path + "/" + directory->name : "";
    if(strPath.find(rootPath) == 0) return fs::u8path(strPath);
    return fs::u8path(rootPath + strPath);
}

std::shared_ptr<file::Directory> FileServer::createDirectory(std::string name, std::shared_ptr<file::Directory> parent) {
    auto path = buildPath(this->rootPath, parent);
    path += name;
    if(!fs::exists(path))
        if(!fs::create_directories(path)) return nullptr;
        else ;
    else if(!fs::is_directory(path)) return nullptr;

    return static_pointer_cast<file::Directory>(this->findFile(path.string()));
}

bool FileServer::fileExists(std::shared_ptr<file::Directory> dir) {
    return fs::exists(buildPath(this->rootPath, dir));
}

bool FileServer::fileExists(std::shared_ptr<file::File> file) {
    return fs::exists(buildPath(this->rootPath, file));
}

std::shared_ptr<file::FileEntry> FileServer::findFile(std::string path, std::shared_ptr<file::Directory> parent) {
	if(path.find(this->rootPath) != 0) {
		string strPath = (parent ? parent->path + "/" + parent->name : this->rootPath) + "/";
		if(strPath.find(this->rootPath) != 0 && rootPath != strPath)
			path = this->rootPath + strPath;
		if(path.find(strPath) == 0 && !strPath.empty())
			;
		else path = strPath + path;
	}

    fs::path absPath = fs::u8path(path);
    if(!fs::is_regular_file(absPath) && !fs::is_directory(absPath)){
        debugMessage(lstream << "Could not find requested file. Abs path: " << absPath << "|" << path << ". (path=" << path << ", parent=" << (parent ? parent->path + "/" + parent->name : "./") << ")");
        return nullptr;
    }

    std::shared_ptr<file::FileEntry> entry;
    if(fs::is_directory(absPath))
        entry = std::make_shared<file::Directory>();
    else entry = std::make_shared<file::File>();

    entry->name = absPath.filename();
    entry->type = fs::is_directory(absPath) ? FileType::DIRECTORY : FileType::FILE;
    entry->path = absPath.parent_path().string();
    entry->lastChanged = fs::last_write_time(absPath);
    if(entry->type == FileType::FILE)
        static_pointer_cast<file::File>(entry)->fileSize = entry->type == FileType::FILE ? fs::file_size(absPath) : 0;
    return entry;
}

std::vector<std::shared_ptr<file::FileEntry>> FileServer::listFiles(std::shared_ptr<file::Directory> dir) {
    if(!dir) return {};
    auto directory = buildPath(this->rootPath, dir);

    if(!fs::exists(directory)) return {};
    if(!fs::is_directory(directory)) return {};

    std::vector<std::shared_ptr<file::FileEntry>> result;
	deque<fs::directory_entry> files;
    for(const auto& elm : fs::directory_iterator(directory)) {
        files.push_back(elm);
    }
	std::sort(files.begin(), files.end(), [](const fs::directory_entry& a, const fs::directory_entry& b) {
		return fs::last_write_time(a.path()) > fs::last_write_time(b.path());
	});
    for(const auto& elm : files) {
        if(fs::is_regular_file(elm.path())){
            auto entry = make_shared<file::File>();
            entry->name = elm.path().filename();
            entry->type = FileType::FILE;
            entry->path = elm.path().parent_path().string();
            entry->lastChanged = fs::last_write_time(elm.path());
            entry->fileSize = fs::file_size(elm.path());
            result.push_back(entry);
        } else if(fs::is_directory(elm.path())){
            auto entry = make_shared<file::Directory>();
            entry->name = elm.path().filename();
            entry->type = FileType::DIRECTORY;
            entry->path = elm.path().parent_path().string();
            entry->lastChanged = fs::last_write_time(elm.path());
            result.push_back(entry);
        } else {
            logError("Invalid file in file tree. File path: " + elm.path().string());
        }
    }

    return result;
}

bool FileServer::deleteFile(std::shared_ptr<file::FileEntry> file) {
    return fs::remove_all(fs::u8path(file->path + "/" + file->name)) > 0;
}

inline std::string randomString(uint length = 15, std::string charIndex = "abcdefghijklmnaoqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ1234567890")
{
    std::string rs = "";
    for (uint i = 0; i < length; ++i)
        rs += charIndex[rand() % charIndex.length()];
    return rs;
}

std::shared_ptr<file::FileTransfereKey> FileServer::generateDownloadTransferKey(std::string &errorMessage, std::string targetFile, size_t offset, const shared_ptr<ConnectedClient>& client) {
    auto file = this->findFile(targetFile, nullptr);
    if(!file){
        errorMessage = "file does not exists";
        return nullptr;
    }
    if(file->type != FileType::FILE) {
        errorMessage = "file is a directory";
        return nullptr;
    }

    auto result = make_shared<file::FileTransfereKey>();
	result->owner = client;
	result->server = client->getServerId();

	result->key_id = rand() % 65535;
    result->createTimestamp = std::chrono::system_clock::now();
    result->key = randomString(16);
    result->offset = offset;
    result->size = static_pointer_cast<file::File>(file)->fileSize;
    result->upload = false;
    result->targetFile = targetFile;

	{ //Test for server?
		auto bandwidth_server = client->getServer()->properties()[property::VIRTUALSERVER_MAX_UPLOAD_TOTAL_BANDWIDTH].as<int64_t>();
		auto bandwidth_instance = serverInstance->properties()[property::SERVERINSTANCE_MAX_UPLOAD_TOTAL_BANDWIDTH].as<int64_t>();

		result->max_server_bandwidth = bandwidth_server;
		result->max_bandwhidth = bandwidth_server;

		if(bandwidth_instance != -1) {
			if(bandwidth_instance < bandwidth_server)
				result->max_bandwhidth = bandwidth_instance; //We want to limit that already until exact calculations was made
		}
	}

	if(client->getServer())
		result->max_server_bandwidth = client->getServer()->properties()[property::VIRTUALSERVER_MAX_DOWNLOAD_TOTAL_BANDWIDTH];

    {
        threads::MutexLock lock(this->keylock);
        pendingKeys.push_back(result);
    }
    return result;
}

std::shared_ptr<file::FileTransfereKey> FileServer::generateUploadTransferKey(std::string &errorMessage, std::string targetFile, size_t size, size_t offset, const shared_ptr<ConnectedClient>& client) {
    shared_ptr<file::FileTransfereKey> result = make_shared<file::FileTransfereKey>();
	result->owner = client;
	result->server = client->getServerId();

	result->key_id = rand() % 65535;
    result->createTimestamp = std::chrono::system_clock::now();
    result->key = randomString(16);
    result->offset = offset;
    result->size = size;
    result->upload = true;
    result->targetFile = targetFile;

	{ //Test for server?
		auto bandwidth_server = client->getServer()->properties()[property::VIRTUALSERVER_MAX_DOWNLOAD_TOTAL_BANDWIDTH].as<int64_t>();
		auto bandwidth_instance = serverInstance->properties()[property::SERVERINSTANCE_MAX_DOWNLOAD_TOTAL_BANDWIDTH].as<int64_t>();

		result->max_server_bandwidth = bandwidth_server;
		result->max_bandwhidth = bandwidth_server;

		if(bandwidth_instance != -1) {
			if(bandwidth_instance < bandwidth_server)
				result->max_bandwhidth = bandwidth_instance; //We want to limit that already until exact calculations was made
		}
	}
    {
        threads::MutexLock lock(this->keylock);
        pendingKeys.push_back(result);
    }
    debugMessage("Created file upload key=" + result->key + " for " + targetFile + " (" + to_string(size) + " bytes)");
    return result;
}

std::shared_ptr<file::Directory> FileServer::resolveDirectory(const shared_ptr<TSServer> &server, std::shared_ptr<BasicChannel> channel, std::string subPath) {
    fs::path path = fs::u8path("server_" + to_string(server ? server->getServerId() : 0) + "/channel_" + to_string(channel->channelId()));
    if(!findFile(path))
        this->createDirectory(path.string(), nullptr);
    path += subPath;
    logMessage(lstream << "resolve " << path.string() << " -> " << findFile(path.string()) << " -> " << typeid(findFile(path.string())).name() << endl);
    return static_pointer_cast<file::Directory>(findFile(path.string()));
}

std::shared_ptr<file::Directory> FileServer::iconDirectory(const shared_ptr<TSServer> &server) {
	fs::path root = fs::u8path(this->rootPath);
    fs::path path = fs::u8path("server_" + to_string(server ? server->getServerId() : 0) + "/icons");
    if(!fs::exists(root / path)) {
	    if(!fs::create_directories(root / path))
	    	return nullptr;
    }
    return static_pointer_cast<file::Directory>(findFile(path.string()));
}

bool FileServer::iconExists(const shared_ptr<TSServer> &server, IconId icon) {
	if(icon == 0) return false;
	if(icon < 1000) return true;
	return this->findFile("icon_" + to_string(icon), this->iconDirectory(server)) != nullptr;
}

std::shared_ptr<file::Directory> FileServer::avatarDirectory(const shared_ptr<TSServer> &server) {
    fs::path path = fs::u8path("server_" + to_string(server ? server->getServerId() : 0) + "/avatars");
    if(!findFile(path))
        this->createDirectory(path.string(), nullptr);
    return static_pointer_cast<file::Directory>(findFile(path.string()));
}

#define TS3_ICON_HASH "icon_1001"
void FileServer::setupServer(const shared_ptr<TSServer> &server) {
    auto dir = iconDirectory(server);
    if(!dir) {
    	logError(LOG_FT,"Failed to find icon directory for server {}", server ? server->getServerId() : 0);
    } else {
	    if(!fs::exists(fs::u8path(dir->path + "/" + dir->name + "/" TS3_ICON_HASH))) {
	    	try {
			    fs::copy(fs::u8path("resources/teaspeak16px.png"), fs::u8path(dir->path + "/" + dir->name + "/" + TS3_ICON_HASH));
	    	} catch(std::exception& ex) {
			    logError(LOG_FT, "Failed to copy default path: {}", ex.what());
	    	}
	    }
    }
}

std::string FileServer::server_file_base(const std::shared_ptr<ts::server::TSServer> &server) {
	return rootPath + "/server_" + to_string(server->getServerId());
}

void FileServer::deleteServer(const shared_ptr<TSServer> &server) {
	fs::path path = fs::u8path(rootPath + "/server_" + to_string(server ? server->getServerId() : 0));
	if(fs::exists(path)) {
		error_code error;
		if(fs::remove_all(path, error) == 0)
			logError(0, "Could not delete server directory {} ({} | {})", path.string(), error.value(), error.message());
	} else {
		logError(0, "Could not delete missing server directory (" + path.string() + ")");
	}
}

//The actual server!
bool FileServer::start(const sockaddr_in& localAdress) {
    if(this->running()) return false;

    this->active = true;
    boundAddress = new sockaddr_in;
    memcpy(boundAddress, &localAdress, sizeof(localAdress));

    this->serverSocket = socket(AF_INET, SOCK_STREAM | SOCK_NONBLOCK, 0);
    if (serverSocket < 0) {
        logCritical(LOG_FT, "Cant create server socket for file server");
        return false;
    }

    int enable = 1;
    int disabled = 0;
    if (setsockopt(serverSocket, SOL_SOCKET, SO_REUSEADDR, &enable, sizeof(int)) < 0) {
	    logError(LOG_FT, "setsockopt(SO_REUSEADDR) failed");
    }
    if(setsockopt(serverSocket, IPPROTO_TCP, TCP_NOPUSH, &disabled, sizeof disabled) < 0) {
	    logError(LOG_FT, lstream << "Cant disable nopush! Error: "+to_string(errno)+" / "+strerror(errno) << endl);
    }
	if(fcntl(serverSocket, F_SETFD, FD_CLOEXEC) < 0) {
		logError(LOG_QUERY, "Failed to enable FD_CLOEXEC for {} (QueryServer)", serverSocket);
	}

    if (bind(serverSocket, (struct sockaddr *) &localAdress, sizeof(localAdress)) < 0) {
        logError(LOG_FT, lstream << "Cant bind server socket (" << strerror(errno) << ")" << endl);
        return false;
    }
    if(listen(serverSocket, 255) < 0){
        logError(LOG_FT, lstream << "Cant listen on server socket (" << strerror(errno) << ")" << endl);
        return false;
    }

    ioLoop = event_base_new();
    this->acceptEvent = event_new(this->ioLoop, this->serverSocket, EV_READ | EV_PERSIST, [](int a, short b, void* c){ ((FileServer*) c)->onClientAccept(a, b, c); }, this);
    event_add(this->acceptEvent, nullptr);

    this->ioThread = new threads::Thread(THREAD_SAVE_OPERATIONS | THREAD_EXECUTE_LATER, [&](){
        event_base_dispatch(this->ioLoop);
        debugMessage(LOG_FT, "File-Server accept thread terminated");
    });
    this->ioThread->name("File IO #1").execute();

    for(int index = 0; index < 2; index++){
        auto th = new threads::Thread(THREAD_SAVE_OPERATIONS | THREAD_EXECUTE_LATER, &FileServer::clientTickingExecutor, this);
        th->name("Ticking FT #" + to_string(index)).execute();
        this->tickingThreads.push_back(th);
    }
    return true;
}

void FileServer::stop() {
    if(!this->running()) return;
    active = false;
	{
		lock_guard<mutex> lock(this->tickingLock);
		this->tickingCon.notify_all();
	}

    if(this->acceptEvent) {
	    event_del(this->acceptEvent);
	    event_free(this->acceptEvent);
	    this->acceptEvent = nullptr;
    }

    auto clClone = this->connectedClients;
    for(const auto& cl : clClone) {
	    cl->disconnect(chrono::milliseconds(1));
    }

	if(this->ioLoop) {
		event_base_loopbreak(this->ioLoop);
		event_base_loopexit(this->ioLoop, nullptr);
	}

	for(const auto& thread : this->tickingThreads){
		if(thread->join(chrono::seconds(1)) != 0) logCritical(LOG_FT, "Failed to terminal file server tick thread!");
		delete thread;
	}
	this->tickingThreads.clear();

    if(this->ioThread)
	    if(this->ioThread->join(chrono::system_clock::now() + chrono::seconds(3)) != 0) {
		    logCritical(LOG_FT, "Could not terminate the event base dispatch thread of the file server");
		    this->ioThread->detach();
	    }
    delete this->ioThread;
    this->ioThread = nullptr;

	if(this->ioLoop) {
		event_base_free(this->ioLoop);
		this->ioLoop = nullptr;
	}

    delete this->boundAddress;
    this->boundAddress = nullptr;

    if(this->serverSocket > 0){
        if(shutdown(this->serverSocket, SHUT_RDWR) < 0) logError(LOG_FT,  "Could not shutdown server socket!");
        if(close(this->serverSocket) < 0) logError(LOG_FT,  "Could not close server socket!");
    }
    this->serverSocket = 0;
}


void FileServer::onClientAccept(int fd, short ev, void *arg) {
    sockaddr_in remoteAddr{};
    memset(&remoteAddr, 0, sizeof(sockaddr_in));
    socklen_t addrLength = sizeof(remoteAddr);

    int acceptedSocketFd = accept(serverSocket, (struct sockaddr *) &remoteAddr, &addrLength);
    if (acceptedSocketFd < 0) {
        if(errno == EAGAIN){ //No manager
	        return;
        }
	    logError(LOG_FT, "Having an error while accepting a new client. ({}/{})", errno, strerror(errno));
        return;
    }

    shared_ptr<FileClient> client = std::make_shared<FileClient>(this, acceptedSocketFd);
    client->_this = client;
    memcpy(&client->remoteAddress, &remoteAddr, sizeof(sockaddr_in));

    this->clientLock.lock();
    this->connectedClients.push_back(client);
    this->clientLock.unlock();
    event_add(client->readEvent, nullptr);
	logMessage(LOG_FT, "Got new client from {}", config::server::disable_ip_saving ? "X.X.X.X" : string(inet_ntoa(client->remoteAddress.sin_addr)) + ":" + to_string(ntohs(client->remoteAddress.sin_port)));
}

void FileServer::clientTickingExecutor() {
    while(this->running()){
        shared_ptr<FileClient> client;
        {
            unique_lock<mutex> lock(this->tickingLock);
            this->tickingCon.wait(lock, [&](){ return !this->running() || !this->tickQueue.empty(); });
            if(!this->running()) return;

            client = this->tickQueue.front();
            this->tickQueue.pop_front();
        }

        if(client->tick()) this->tickFileClient(client); //Client needs more ticking
    }
}

std::deque<std::shared_ptr<FileClient>> FileServer::running_file_transfers(const std::shared_ptr<ts::server::ConnectedClient> &client) {
	std::deque<std::shared_ptr<FileClient>> result;

	{
		threads::MutexLock lock(this->clientLock);
		for(const auto& c : this->connectedClients) {
			if(!client || c->client == client) result.push_back(c);
		}
	}

	return result;
}

std::deque<std::shared_ptr<file::FileTransfereKey>> FileServer::pending_file_transfers(const std::shared_ptr<ts::server::ConnectedClient> &client) {
	std::deque<std::shared_ptr<file::FileTransfereKey>> result;
	for(const auto& key : this->pending_keys())
		if(!client || key->owner.lock() == client)
			result.push_back(key);
	return result;
}

void FileServer::tickFileClient(std::shared_ptr<FileClient> cl) {
    lock_guard<mutex> lock(this->tickingLock);
    this->tickQueue.push_back(cl);
    this->tickingCon.notify_one();
}

struct TransfareGroup {
	shared_ptr<TSServer> server;
	deque<shared_ptr<FileClient>> clients;
	enum {
		upload,
		download
	} direction = upload;
	size_t used_bandwidth = 0;
	ssize_t max_bandwidth = -1;
};

void FileServer::instanceTick() {
	{
		//tickQueue
		auto client = this->connected_clients();
		lock_guard<mutex> tick_lock(this->tickingLock);
		this->tickQueue.insert(this->tickQueue.end(), client.begin(), client.end()); //Tick all clients :)
		this->tickingCon.notify_all();
	}

	auto now = system_clock::now();
	if(timestamp_bandwidth_update + seconds(1) < now) {
		timestamp_bandwidth_update = now;

		for(const auto& entry : this->pending_keys()) {
			if(!entry) continue;
			if(entry->createTimestamp + minutes(1) < now) {
				logMessage(entry->server, "[FILE] Timeout file transfer for file " + entry->targetFile + "! Dropping key");

				{
					threads::MutexLock lock(this->keylock);
					auto index = find(this->pendingKeys.begin(), this->pendingKeys.end(), entry);
					if(index != this->pendingKeys.end())
						this->pendingKeys.erase(index);
				}
			}
		}

		deque<shared_ptr<TransfareGroup>> groups;
		{
			for(const auto& client : this->connected_clients()) {
				if(!client || !client->client) continue;

				lock_guard<threads::Mutex> l(client->tickLock); //Instance hangup here
				if(!client->pendingKey) continue;

				auto used_band = client->used_bandwidth();
				if(client->pendingKey && client->state_connection != FileClient::C_DISCONNECTED && (client->pendingKey->max_bandwhidth < 0 || used_band < client->pendingKey->max_bandwhidth)) {

					if(client->event_read_hold) { //Reactivate read!
						event_add(client->readEvent, nullptr);
						client->event_read_hold = false;
						debugMessage(LOG_FT, "{} Reattaching client read event! (Current bandwidth {})", client->client_prefix(), used_band);
					}
					if(client->event_write_hold) { //Reactivate read!
						event_add(client->writeEvent, nullptr);
						client->event_write_hold = false;
						debugMessage(LOG_FT, "{} Reattaching client write event! (Current bandwidth {})", client->client_prefix(), used_band);
					}
				}
				if(!client->client->getServer()) continue;

				bool found = false;
				for(auto& group : groups) {
					if (group->server == client->client->getServer() && (group->direction == TransfareGroup::upload) == client->pendingKey->upload) {
						group->clients.push_back(client);
						group->used_bandwidth += client->used_bandwidth();
						found = true;
						break;
					}
				}
				if(found) continue;

				auto entry = make_shared<TransfareGroup>();
				entry->direction = client->pendingKey->upload ? TransfareGroup::upload : TransfareGroup::download;
				entry->used_bandwidth = client->used_bandwidth();
				entry->clients.push_back(client);
				entry->server = client->client->getServer();
				groups.push_back(entry);
			}
		}

		{ //Instance limit upload
			auto limit_upload = serverInstance->properties()[property::SERVERINSTANCE_MAX_UPLOAD_TOTAL_BANDWIDTH].as<ssize_t>();
			if(limit_upload >= 0) {
				deque<shared_ptr<TransfareGroup>> elements;
				size_t used_bandwidth = 0;
				for(const auto& entry : groups) {
					if (entry->direction == TransfareGroup::upload) {
						elements.push_back(entry);
						used_bandwidth += entry->used_bandwidth;
					}
				}
				if(!elements.empty() && used_bandwidth > limit_upload) {
					for(auto& entry : elements) {
						entry->max_bandwidth = ssize_t(ceil(double(entry->used_bandwidth) * double(limit_upload) / double(used_bandwidth))); //Adjust the total
					}
				}
			}
		}

		{ //Instance limit download
			auto limit_download = serverInstance->properties()[property::SERVERINSTANCE_MAX_DOWNLOAD_TOTAL_BANDWIDTH].as<ssize_t>();
			if(limit_download >= 0) {
				deque<shared_ptr<TransfareGroup>> elements;
				size_t used_bandwidth = 0;
				for(const auto& entry : groups) {
					if (entry->direction == TransfareGroup::download) {
						elements.push_back(entry);
						used_bandwidth += entry->used_bandwidth;
					}
				}
				if(!elements.empty() && used_bandwidth > limit_download) {
					for(auto& entry : elements) {
						entry->max_bandwidth = ssize_t(ceil(double(entry->used_bandwidth) * double(limit_download) / double(used_bandwidth))); //Adjust the total
					}
				}
			}
		}

		map<ServerId, deque<shared_ptr<TransfareGroup>>> servers;
		for(const auto& entry : groups) {
			servers[entry->server->getServerId()].push_back(entry);
		}

		for(const auto& server : servers) {
			if(server.second.empty()) continue;

			{ //Adjust server upload
				auto max_bandwidth = server.second.front()->server->properties()[property::VIRTUALSERVER_MAX_UPLOAD_TOTAL_BANDWIDTH].as<ssize_t>();
				if(max_bandwidth >= 0) {
					deque<shared_ptr<TransfareGroup>> tranfares;
					size_t bandwidth = 0;
					for(const auto& trans : server.second) {
						if(trans->direction == TransfareGroup::upload) {
							tranfares.push_back(trans);
							bandwidth += trans->max_bandwidth;
						}
					}

					if(bandwidth > max_bandwidth) {
						for(auto& trans : tranfares)
							trans->max_bandwidth = ssize_t(ceil(double(trans->max_bandwidth) * double(max_bandwidth) / double(bandwidth))); //Adjust the total
					}
				}
			}

			{ //Adjust server download
				auto max_bandwidth = server.second.front()->server->properties()[property::VIRTUALSERVER_MAX_DOWNLOAD_TOTAL_BANDWIDTH].as<ssize_t>();
				if(max_bandwidth >= 0) {
					deque<shared_ptr<TransfareGroup>> tranfares;
					size_t bandwidth = 0;
					for(const auto& trans : server.second)
						if(trans->direction == TransfareGroup::download) {
							tranfares.push_back(trans);
							bandwidth += trans->max_bandwidth;
						}

					if(bandwidth > max_bandwidth) {
						for(auto& trans : tranfares)
							trans->max_bandwidth = ssize_t(ceil(double(trans->max_bandwidth) * double(max_bandwidth) / double(bandwidth))); //Adjust the total
					}
				}
			}
		}
	}
}