#pragma once

#include <vector>
#include <string>
#include <chrono>
#include <memory>
#include <ThreadPool/Thread.h>
#include <deque>
#include <ThreadPool/Mutex.h>
#include <netinet/in.h>
#include <event.h>
#include <condition_variable>
#include "Variable.h"
#include <Definitions.h>

namespace ts {
    namespace file {
        namespace FileType {
            enum FileType {
                DIRECTORY,
                FILE
            };
        }
    }
}
DEFINE_VARIABLE_TRANSFORM(ts::file::FileType::FileType, VARTYPE_INT, std::to_string((uint8_t) in), static_cast<ts::file::FileType::FileType>(in.as<uint8_t>()));
namespace ts {
    class BasicChannel;

	namespace server {
		class ConnectedClient;
	}

    namespace file {
        struct FileEntry {
            FileType::FileType type;
            std::string path;
            std::string name;
            std::chrono::time_point<std::chrono::system_clock> lastChanged;
        };

        struct File : public FileEntry {
            int64_t fileSize;
        };

        struct Directory : public FileEntry { };


        struct SymLink : public FileEntry {
            std::string targetPath;
            std::string targetName;
        };

        struct FileTransfereKey {
        	uint16_t key_id = 0;
	        ServerId server;
	        std::weak_ptr<server::ConnectedClient> owner;

            std::chrono::time_point<std::chrono::system_clock> createTimestamp;
            std::string key;
            std::string targetFile; //Relative
            int64_t size;
            size_t offset;

            bool upload = false;

	        int64_t max_bandwhidth = -1; //Get calculated each time
	        int64_t max_server_bandwidth = -1;
        };
    }


    namespace server {
        class TSServer;
        class FileClient;

        //FIXME Valid path
        class FileServer {
                friend class FileClient;
            public:
                FileServer();
                ~FileServer();

                bool start(const sockaddr_in&);
                void stop();
                bool running(){ return active; }
                sockaddr_in* boundedAddress(){ return boundAddress; }

                std::shared_ptr<file::Directory> createDirectory(std::string name, std::shared_ptr<file::Directory> parent);
                bool fileExists(std::shared_ptr<file::Directory>);
                bool fileExists(std::shared_ptr<file::File>);
                std::shared_ptr<file::FileEntry> findFile(std::string, std::shared_ptr<file::Directory> = nullptr);
                std::vector<std::shared_ptr<file::FileEntry>> listFiles(std::shared_ptr<file::Directory> directory = nullptr);
                bool deleteFile(std::shared_ptr<file::FileEntry>);


                std::shared_ptr<file::FileTransfereKey> generateDownloadTransferKey(std::string &errorMessage, std::string targetFile, size_t offset, const std::shared_ptr<ConnectedClient>&);
                std::shared_ptr<file::FileTransfereKey> generateUploadTransferKey(std::string &errorMessage, std::string targetFile, size_t size, size_t offset, const std::shared_ptr<ConnectedClient>&);

                std::shared_ptr<file::Directory> resolveDirectory(const std::shared_ptr<TSServer> &, std::shared_ptr<BasicChannel>, std::string = "");
                std::shared_ptr<file::Directory> iconDirectory(const std::shared_ptr<TSServer> &);
		        bool iconExists(const std::shared_ptr<TSServer> &, IconId);
                std::shared_ptr<file::Directory> avatarDirectory(const std::shared_ptr<TSServer> &);

                std::string server_file_base(const std::shared_ptr<TSServer> &);
                void setupServer(const std::shared_ptr<TSServer> &);
		        void deleteServer(const std::shared_ptr<TSServer> &);

                void tickFileClient(std::shared_ptr<FileClient>);
                void instanceTick();

		        std::deque<std::shared_ptr<FileClient>> running_file_transfers(const std::shared_ptr<ConnectedClient> & /* client */ = nullptr);
		        std::deque<std::shared_ptr<file::FileTransfereKey>> pending_file_transfers(const std::shared_ptr<ConnectedClient> & /* client */ = nullptr);
            private:
                sockaddr_in* boundAddress = nullptr;;

                bool active = false;
                int serverSocket;
                std::string rootPath = "./files/";

                //IO stuff
                event_base* ioLoop = nullptr;
                ::event* acceptEvent = nullptr;

                threads::Mutex clientLock;
                std::deque<std::shared_ptr<FileClient>> connectedClients;
		        inline std::deque<std::shared_ptr<FileClient>> connected_clients(){
			        threads::MutexLock lock(clientLock);
			        return connectedClients;
		        }

                threads::Thread* ioThread = nullptr;
                void onClientAccept(int fd, short ev, void *arg);

                std::deque<std::shared_ptr<FileClient>> tickQueue;
                std::deque<threads::Thread*> tickingThreads;
                std::mutex tickingLock;
                std::condition_variable tickingCon;

                std::chrono::system_clock::time_point timestamp_bandwidth_update;

                //file management
                threads::Mutex keylock;
                std::deque<std::shared_ptr<file::FileTransfereKey>> pendingKeys;
                inline std::deque<std::shared_ptr<file::FileTransfereKey>> pending_keys(){
                	threads::MutexLock lock(keylock);
                	return pendingKeys;
                }

                void clientTickingExecutor();
        };
    }
}